#include "consola.h"
#include "hilos.h"
#include "lista.h"
#include "varios.h"

#define MAXBUFTCL		256
#define VIDEOMONO		0xB000
#define VIDEOCOLOR	0xB800
#define DIR6845MONO	0x03B4;
#define DIR6845COLOR	0x03D4;
#define NUMFILA		25
#define NUMCOLUMNA	80

TLista tlHilosBloqueados;

static int far *nPantalla;
static int nDir6845;
static int nFila=0;
static int nColumna=0;
static char cBufferTeclado[MAXBUFTCL];
static int nInicioLista=0; /* Cabeza de la lista. Vacia si Inicio == Final */
static int nFinalLista=0; /* Llena si (Final+1)%MAXBUFTCL == Inicio */
static char cTecladoNormal[128]={
0x00,0x1b,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x30,0x27,0x3d,0x7d,0x09,
0x71,0x77,0x65,0x72,0x74,0x79,0x75,0x69,0x6f,0x70,0x60,0x2b,0x0d,0x00,0x61,0x73,
0x64,0x66,0x67,0x68,0x6a,0x6b,0x6c,0x3b,0x2c,0xaa,0x00,0x7d,0x7a,0x78,0x63,0x76,
0x62,0x6e,0x6d,0x2c,0x2e,0x2d,0x00,0x2a,0x00,0x20,0x00,0x3b,0x3c,0x3d,0x3e,0x3f,
0x40,0x41,0x42,0x43,0x44,0x00,0x00,0x47,0x48,0x49,0x2d,0x4b,0x00,0x4d,0x2b,0x4f,
0x50,0x51,0x52,0x53,0x00,0x00,0x3c,0x85,0x86,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};

void interrupt ManejadorTeclado(void)
{
	unsigned char ucodigo;
	int id;

	while ((LeerBytePuerto(0x64) & 0x01) != 0) {
		ucodigo = LeerBytePuerto(0x60);
		if ((ucodigo < 128) &&
			(((nFinalLista + 1) % MAXBUFTCL) != nInicioLista)) {
			if (cTecladoNormal[ucodigo] != '\0') {
				cBufferTeclado[nFinalLista] = cTecladoNormal[ucodigo];
				nFinalLista = (nFinalLista + 1) % MAXBUFTCL;
				id = ExtraerHilo(&tlHilosBloqueados);
				if (id >= 0) {
					thTablaHilos[id].estado = LISTO;
					IntroducirHilo(id, &tlHilosPreparados);
				}
			}
		}
	}
	EscribirBytePuerto(0x20, 0x20);
}

char MHLeerCaracter(void)
{
	int nflags;
	char ccaracter;

	nflags = DesactivarInt();
	while (nInicioLista == nFinalLista) { /* Ningun caracter disponible */
		thTablaHilos[nHiloActual].estado = BLOQUEADO;
		IntroducirHilo(nHiloActual, &tlHilosBloqueados);
		Planificador();
		ConmutarContexto();
	}
	ccaracter = cBufferTeclado[nInicioLista];
	nInicioLista = (nInicioLista + 1) % MAXBUFTCL;
	RecuperarFlags(nflags);
	return ccaracter;
}

static void PosicionarCursor(int nFila, int nColumna)
{
	int nposicion;

	nposicion = nFila * NUMCOLUMNA + nColumna;
	EscribirBytePuerto(nDir6845, 14);
	EscribirBytePuerto(nDir6845 + 1, nposicion >> 8);
	EscribirBytePuerto(nDir6845, 15);
	EscribirBytePuerto(nDir6845 + 1, nposicion);
}

static void LeerPosicionCursor(void)
{
	int nposicion;
	int nbyte;

	EscribirBytePuerto(nDir6845, 14);
	nbyte = LeerBytePuerto(nDir6845 + 1);
	nposicion = nbyte * 256;
	EscribirBytePuerto(nDir6845, 15);
	nbyte = LeerBytePuerto(nDir6845 + 1);
	nposicion += nbyte;
	nFila = nposicion / NUMCOLUMNA;
	nColumna = nposicion % NUMCOLUMNA;
}

static void NuevaLinea(void)
{
	int i, nnumero;

	nColumna = 0;
	if (nFila < (NUMFILA - 1)) nFila++;
	else {
		nFila = NUMFILA - 1;
		/* Codigo del scroll */
		nnumero = NUMFILA * NUMCOLUMNA;
		for (i=0; i<nnumero; i++) {
			*(nPantalla + i) = *(nPantalla + i + NUMCOLUMNA);
		}
		for (i=nnumero; i<nnumero+NUMCOLUMNA; i++) *(nPantalla + i) = 0x0720;
	}
}

int MHEscribirCaracter(char c)
{
	int nflags;

	nflags = DesactivarInt();
	if (c == '\n') NuevaLinea();
	else {
		/* Escritura del caracter */
		*(nPantalla + nFila*NUMCOLUMNA + nColumna) = c + 0x0700;
		/* Actualizacion de la posicion del cursor */
		nColumna = (nColumna + 1) % NUMCOLUMNA;
		if (nColumna == 0) NuevaLinea();
	}
	PosicionarCursor(nFila, nColumna);
	RecuperarFlags(nflags);
	return 0;
}

int MHEscribirCadena(const char far *c)
{
	int nflags;

	nflags = DesactivarInt();
	while (*c != '\0') MHEscribirCaracter(*c++);
	RecuperarFlags(nflags);
	return 0;
}

void EscribirNumero(int nnumero)
{
	int nflags, nresto, i, nindice=0;
	char cbuffer[10];

	nflags = DesactivarInt();
	if (nnumero == 0) {
		MHEscribirCadena("0 ");
		return;
	}
	if (nnumero < 0) {
		MHEscribirCaracter('-');
		nnumero =  -nnumero;
	}
	while (nnumero != 0) {
		nresto = nnumero % 10;
		cbuffer[nindice++] = nresto + '0';
		nnumero = nnumero / 10;
	}
	for (i=nindice - 1; i >= 0; i--) MHEscribirCaracter(cbuffer[i]);

	RecuperarFlags(nflags);
}

void EscribirCadenaNumero(const char *c, int nnumero)
{
	int nflags;

	nflags = DesactivarInt();
	MHEscribirCadena(c);
	EscribirNumero(nnumero);
	RecuperarFlags(nflags);
}

void InicializacionConsola(void)
{
	int far *nhardware;
	/* Tipo de tarjeta grafica */
	nhardware = MK_FP(0, 0x0410);
	if ((*nhardware & 0x0030) == 0x0030) { /* Monocromo */
		nPantalla = MK_FP(VIDEOMONO, 0);
		nDir6845 = DIR6845MONO;
	}
	else { /* Color */
		nPantalla = MK_FP(VIDEOCOLOR, 0);
		nDir6845 = DIR6845COLOR;
	}

	/* Posicion inicial del cursor */
	LeerPosicionCursor();
}
